"""
08/21, R James, F Alder

Adapted from script by https://github.com/awmlee
"""
from websocket import create_connection
import requests
import json
import keyboard


class Websocket:
    def __init__(self, socket_url=None):
        self.rp_ip = 'rp-f05a98.local' if socket_url == '' else socket_url
        self.socket_url = 'ws://' + self.rp_ip + '/wss'
        self.start_app_url = 'http://' + self.rp_ip + '/bazaar?start=streaming_manager'



#------------------------------------------------------------------------------
socket_url = input('Red Pitaya IP address (press enter for default: rp-f05a98.local):')
WS = Websocket(socket_url)

# Start streaming server
try:
    web = requests.get(WS.start_app_url)
    print('Connection established')
except:
    raise OSError('Could not connect to Red Pitaya')

# Create websocket connection
ws = create_connection(WS.socket_url)

# Set websocket parameters
setbits = ({'parameters': {'SS_RESOLUTION': {'value': 2}, 'in_command': {'value': 'send_all_params'}}})
setrate = ({'parameters': {'SS_RATE': {'value': 800}, 'in_command': {'value': 'send_all_params'}}})
setdualchan = ({'parameters': {'SS_CHANNEL': {'value': 3},' in_command': {'value': 'send_all_params'}}})
ws.send(json.dumps(setbits))
ws.send(json.dumps(setrate))
ws.send(json.dumps(setdualchan))

startcmd = ({'parameters': {'SS_START': {'value': 1}, 'in_command': {'value': 'send_all_params'}}})
stopcmd = ({'parameters': {'SS_START': {'value': 0}, 'in_command': {'value': 'send_all_params'}}})

# Start streaming data
ws.send(json.dumps(startcmd))

print('Streaming: press q to stop')
while True:
    if keyboard.read_key() == 'q':
        print('Ending streaming, closing connection')
        break

# Stop streaming data and close websocket
ws.send(json.dumps(stopcmd))
ws.close()



"""
08/21, R James, F Alder

Adapted from script by https://github.com/awmlee
"""

import numpy as np
import time
import os

import PyRPStream as rp
export, __all__ = rp.exporter()


@export
class RPDevice:
    """
    """
    def __init__(self, name):
        # Device name
        self.name = name
        # Calibration parameters
        if os.path.isfile(self.name + '_calib.txt'):
            file = open(self.name + '_calib.txt', 'r')
            calib_consts = file.readlines()
            self.ch1_offset = float(calib_consts[0])
            self.ch1_gain = float(calib_consts[1])
            self.ch2_offset = float(calib_consts[2])
            self.ch2_gain = float(calib_consts[3])
        else:
            self.ch1_offset = 0.
            self.ch1_gain = 1.
            self.ch2_offset = 0.
            self.ch2_gain = 1.
        # Connection information
        self.address_port = ('rp-f05a98.local', 8900)
        # Device information
        self.input_range_V = 2.
        self.input_bits = 16
        # Socket thread
        self.client = rp.SocketClientThread()
        # Start the socket thread
        self.client.start()


    def connect(self):
        """
        """
        if not self.client.alive.isSet():
            # Create new socket thread
            self.client = rp.SocketClientThread()
            # Start the socket thread
            self.client.start()

        print('Trying to CONNECT to socket')
        # Make 10 attempts, then give up
        for i in range(10):
            self.client.cmd_q.put(rp.ClientCommand('CONNECT', self.address_port))
            client_reply = self.client.reply_q.get()

            if client_reply.key == 'ERROR':
                print(client_reply.reply)
                print('Trying to CONNECT again')
            elif client_reply.key == 'MESSAGE':
                print(client_reply.reply)
                break
            else:
                break

        # Check if we are connected by attempting to RECEIVE
        self.client.cmd_q.put(rp.ClientCommand('RECEIVE'))
        client_reply = self.client.reply_q.get()
        if client_reply.key == 'ERROR':
            # We aren't connected: end the thread
            print(client_reply.reply)
            self.client.join()
            raise OSError('Repeatedly failed to execute CONNECT: exiting')

        time.sleep(1)


    def disconnect(self):
        """
        """
        print('Trying to CLOSE socket connection')

        if not self.client.alive.isSet():
            # There is no thread
            return

        if not self.client.connected:
            # We aren't connected: end the thread
            self.client.join()
            return

        self.client.cmd_q.put(rp.ClientCommand('CLOSE'))

        time.sleep(1)

        # Get reply upon socket closure
        client_reply = self.client.reply_q.get()
        print(client_reply.reply)

        # End the thread
        self.client.join()


    def acquire(self, acq_time, file_size=1e6):
        """
        """
        if not self.client.alive.isSet():
            # There is no thread
            raise OSError('Cannot acquire when not connected to device')

        if not self.client.connected:
            # We aren't connected: end the thread
            self.client.join()
            raise OSError('Cannot acquire when not connected to device')

        try:
            assert(acq_time > 0)
        except:
            raise TypeError('Invalid acquisition time value')

        # Number of messages to use for data rate calculation
        n_samp = 10

        # Get reply from reply queue to obtain start time
        client_reply = self.client.reply_q.get()

        if client_reply.key == 'ERROR':
            # End the thread if we receive ERROR
            print(client_reply.reply)
            self.client.join()
            raise OSError('Error during RECEIVE: exiting')

        elif client_reply.key == 'DATA':
            # If we have DATA, get the start acquisition time
            t_start = client_reply.reply['timestamp']

        t_prev = t_start
        i = 0

        ch1_data = bytearray()
        ch2_data = bytearray()
        ch1_reads = 0
        ch2_reads = 0
        t_file_ch1 = None
        t_file_ch2 = None

        # Loop to RECEIVE DATA, unless an ERROR occurs
        while True:
            # Get reply from reply queue
            client_reply = self.client.reply_q.get()

            if client_reply.key == 'ERROR':
                # End the thread if we receive ERROR
                print(client_reply.reply)
                self.client.join()
                raise OSError('Error during RECEIVE: exiting')

            elif client_reply.key == 'DATA':
                # If we have DATA, exit if we have exceeded acquisition time
                t = client_reply.reply['timestamp']
                if t - t_start > acq_time:
                    break

                # Otherwise,
                if ch1_reads == 0:
                    t_file_ch1 = t
                if ch2_reads == 0:
                    t_file_ch2 = t

                ch1_data += client_reply.reply['ch1_data']
                ch2_data += client_reply.reply['ch2_data']
                ch1_reads += 1
                ch2_reads += 1

                if (ch1_reads * self.client.ch1_size > file_size):
                    f=open(f'red_pitaya_data_ch1_{t_file_ch1}.bin', 'wb')
                    f.write(ch1_data)
                    f.close()
                    ch1_data = bytearray()
                    ch1_reads = 0
                if (ch2_reads * self.client.ch2_size > file_size):
                    f=open(f'red_pitaya_data_ch2_{t_file_ch2}.bin', 'wb')
                    f.write(ch2_data)
                    f.close()
                    ch2_data = bytearray()
                    ch2_reads = 0

                # Calculate the data rate every n_samp messages
                if i == n_samp:
                    print('Data rate is ', n_samp * 2**16 / (t - t_prev) / 1024 / 1024)  # MBytes/second
                    i = 0
                    t_prev = t

                i += 1


    def acquire_calib(self, acquire_raw=False):
        """
        """
        if not self.client.alive.isSet():
            # There is no thread
            raise OSError('Cannot calibrate when not connected to device')

        if not self.client.connected:
            # We aren't connected: end the thread
            self.client.join()
            raise OSError('Cannot calibrate when not connected to device')

        # Get reply from reply queue to obtain start time
        client_reply = self.client.reply_q.get()

        if client_reply.key == 'ERROR':
            # End the thread if we receive ERROR
            print(client_reply.reply)
            self.client.join()
            raise OSError('Error during RECEIVE: exiting')

        elif client_reply.key == 'DATA':
            # If we have DATA, get the start acquisition time
            t_start = client_reply.reply['timestamp']

        ch1_data_store = []
        ch2_data_store = []

        # Loop to RECEIVE DATA, unless an ERROR occurs
        while True:
            # Get reply from reply queue
            client_reply = self.client.reply_q.get()

            if client_reply.key == 'ERROR':
                # End the thread if we receive ERROR
                print(client_reply.reply)
                self.client.join()
                raise OSError('Error during RECEIVE: exiting')

            elif client_reply.key == 'DATA':
                # If we have DATA, exit if we have exceeded acquisition time
                t = client_reply.reply['timestamp']
                if t - t_start > 1: # Acquire for 1 second
                    break

                # Otherwise, append data
                ch1_data = np.frombuffer(client_reply.reply['ch1_data'], dtype=np.int16)
                ch2_data= np.frombuffer(client_reply.reply['ch2_data'], dtype=np.int16)

                if not acquire_raw:
                # Convert from ADC -> V, with calibration factors, by default
                    ch1_data = self.ch1_gain * (ch1_data * self.input_range_V / 2. ** self.input_bits + self.ch1_offset)
                    ch2_data = self.ch2_gain * (ch2_data * self.input_range_V / 2. ** self.input_bits + self.ch2_offset)

                ch1_data_store = np.append(ch1_data_store, ch1_data)
                ch2_data_store = np.append(ch2_data_store, ch2_data)

        # Save calibration data to files
        np.savetxt('red_pitaya_data_ch1_calib.txt', ch1_data_store)
        np.savetxt('red_pitaya_data_ch2_calib.txt', ch2_data_store)



"""
09/21, R James, F Alder
"""

import numpy as np
import time
import os

import PyRPStream as rp
export, __all__ = rp.exporter()


@export
class BinaryDecoder:
    """
    """
    def __init__(self, name, ignore_calib=False):
        # Device name
        self.name = name
        # Calibration parameters
        if os.path.isfile(self.name + '_calib.txt') and not ignore_calib:
            file = open(self.name + '_calib.txt', 'r')
            calib_consts = file.readlines()
            self.ch1_offset = float(calib_consts[0])
            self.ch1_gain = float(calib_consts[1])
            self.ch2_offset = float(calib_consts[2])
            self.ch2_gain = float(calib_consts[3])
        else:
            self.ch1_offset = 0
            self.ch1_gain = 1
            self.ch2_offset = 0
            self.ch2_gain = 1


    def decode_file(channel, timestamp):
        """
        """
        try:
            assert(channel == 1 or 2)
        except:
            raise ValueError('Invalid channel value (must be 1 or 2)')

        try:
            assert(timestamp > 0)
        except:
            raise TypeError('Invalid timestamp value')

        if channel == 1:
            filename = 'red_pitaya_data_ch1_' + f'{timestamp}' + '.bin'
        elif channel == 2:
            filename = 'red_pitaya_data_ch2_' + f'{timestamp}' + '.bin'

        try:
            assert os.path.isfile(filename)
        except:
            raise OSError('Data at specified timestamp and channel does not exist')

        data_btyes = np.loadtxt(filename)

        data_decoded = np.frombuffer(data_bytes, dtype=np.int16)

        print(data_decoded)
        # ch1_data_calib = self.ch1_gain * (ch1_data_no_calib  * self.input_range_V / 2 ** self.input_bits + self.ch1_offset)



__version__ = '0.8.5'
__author__ = 'Robert James, Fiona Alder'
__author_email__ = 'robert.james.19@ucl.ac.uk'

from .utils import *
from .socket_thread import *
from .device_interface import *
from .calibration import *



"""
08/21, R James, F Alder

Adapted from script by https://github.com/awmlee
"""

import socket
import threading
import queue
import numpy as np
import time

import PyRPStream as rp
export, __all__ = rp.exporter()


@export
class ClientCommand:
    """ A command to the client thread.
        Each command key has its associated command type:

        'CONNECT':    (host, port) tuple
        'RECEIVE':    None
        'CLOSE':      None
    """
    def __init__(self, key, command=None):
        self.key = key
        self.command = command



@export
class ClientReply:
    """ A reply from the client thread.
        Each reply key has its associated reply type:

        'ERROR':      The error string
        'DATA':       Depends on the command - for RECEIVE it's the received
                      data string, for others None
        'MESSAGE':    Status message
    """
    def __init__(self, key, reply=None):
        self.key = key
        self.reply = reply



@export
class SocketClientThread(threading.Thread):
    """ Implements the threading (run, join, etc.): the thread can be
        controlled via the cmd_q queue attribute. Replies are placed in
        the reply_q queue attribute.
    """
    def __init__(self):
        super(SocketClientThread, self).__init__()

        # Command and reply queues for communicating with the socket thread
        self.cmd_q = queue.Queue()
        self.reply_q = queue.Queue()
        # Thread run control
        self.alive = threading.Event()
        self.alive.set()
        # Socket connection attributes
        self.connected = 0
        self.socket = None
        # Buffer reading attributes
        self.header_size = 60
        self.ch1_size = 32768
        self.ch2_size = 32768
        # Data storage
        self.data_all_ch1 = []
        self.data_all_ch2 = []
        # Thread control handlers
        self.handlers = {
            'CONNECT': self._handle_CONNECT,
            'RECEIVE': self._handle_RECEIVE,
            'CLOSE': self._handle_CLOSE
        }

        print('Starting socket thread')


    def run(self):
        """ Thread control function.
        """
        while self.alive.isSet():
            if self.connected:
                # Default is to RECEIVE
                command = ClientCommand('RECEIVE')
                self.handlers[command.key](command)
            else:
                # Don't do anything if not connected
                time.sleep(0.1)
            try:
                # Check if any commands have been sent to the thread
                command = self.cmd_q.get(block=False)
                self.handlers[command.key](command)
            except queue.Empty:
                # Do nothing if no commands sent
                continue


    def join(self, timeout=None):
        """ Invoking this will end the thread.
        """
        print('Ending socket thread')

        self.alive.clear()
        threading.Thread.join(self, timeout)


    def _handle_CONNECT(self, client_command):
        """ Attempt to connect to the socket, if not already connected
        client_command should be a (host, port) tuple.
        """
        try:
            if self.connected == 0:
                # Attempt to connect if not already connected
                self.socket = socket.socket(
                    socket.AF_INET, socket.SOCK_STREAM)
                self.socket.connect((client_command.command[0], client_command.command[1]))
                self.reply_q.put(self._message_reply('Socket connected'))
                self.connected = 1
            else:
                # We are already connected, do nothing
                self.reply_q.put(self._message_reply('Socket already connected'))
        except OSError as e:
            # Return error if we can't connect
            self.reply_q.put(self._error_reply(str(e)))
            self.connected = 0


    def _handle_RECEIVE(self, client_command):
        """ Read from the socket: discard the header information, then send
        data from channel 1 and channel 2 to the reply queue.
        """
        try:
            reply = {}

            # Take the header information from the socket, then discard
            header_bytes = self._receieve_bytes(self.header_size, self.socket)

            # Take channel 1 data from the socket, store in reply
            ch1_bytes = self._receieve_bytes(self.ch1_size, self.socket)
            reply.update({'ch1_data': ch1_bytes})

            # Take channel 2 data from the socket, store in reply
            ch2_bytes = self._receieve_bytes(self.ch2_size, self.socket)
            reply.update({'ch2_data': ch2_bytes})

            # Store the timestamp in reply
            reply.update({'timestamp': time.time()})

            # Put reply in the reply queue
            self.reply_q.put(self._data_reply(reply), block=True)

        except OSError as e:
            self.reply_q.put(self._error_reply(str(e)))
            self.connected = 0


    def _receieve_bytes(self, n, socket):
        """ Receive n bytes from socket.
        """
        tdata = bytearray()
        while len(tdata) < n:
            packet = socket.recv(n - len(tdata))
            if not packet:
                raise OSError('Empty packet from socket.recv()')
            tdata.extend(packet)

        return tdata


    def _handle_CLOSE(self, client_command):
        """ Close connection to the socket.
        """
        self.socket.close()
        self.reply_q.queue.clear()
        self.reply_q.put(self._message_reply('Socket closed'))
        self.connected = 0


    def _error_reply(self, errstr):
        return ClientReply('ERROR', errstr)


    def _message_reply(self, data=None):
        return ClientReply('MESSAGE', data)


    def _data_reply(self, data=None):
        return ClientReply('DATA', data)
